#
#  This module has all the code for trace profile application
#  and calculation of centroids for reference profile and science
#  data - the TRCECORR and ALGNCORR steps
#

from astropy.io import fits
import numpy as np
import math

from . import airglow
from .calcosparam import *
from . import ccos
from . import cosutil
from . import dispersion

TRACE_OK = 1
CENTROID_OK = 1
CENTROID_UNDEFINED = 2
CENTROID_ERROR_TOO_LARGE = 3
CENTROID_SET_BY_USER = 4
NO_CONVERGENCE = 5

#
# Temporarily use one of the unused bits in the DQ array to mark
# airglow lines

DQ_AIRGLOW = 1024

#
# Minimum number of good columns to make a row 'good'

COLMIN = 100

def doTrace(events, info, reffiles, tracemask):
    """Do the trace correction.  The trace reference file follows the
    centroid of a point source.  Applying the correction involves
    subtracting the profile of trace vs. xcorr from yfull"""
    traceprofile = getTrace(reffiles['tracetab'], info)
    #
    # Remove bad values from the trace
    cleanTrace(traceprofile)
    #
    # Apply the trace correction to the yfull values
    applyTrace(events.field('xcorr'), events.field('yfull'), traceprofile,
               tracemask)
    
    return traceprofile

def getTrace(tracetab, info):
    """Get the trace from the tracetab reference file."""
    filter = {"segment": info["segment"],
              "opt_elem": info["opt_elem"],
              "cenwave": info["cenwave"],
              "aperture": info["aperture"]
              }
    trace_row = cosutil.getTable(tracetab, filter)
    if trace_row is None:
        raise MissingRowError("Missing row in TRACETAB, filter = %s" %  str(filter))
    trace = trace_row["TRACE"][0]
    return trace

def cleanTrace(trace):
    """Clean the trace.  For now, this just means replacing values
    of -999.0 with 0.0"""
    bad = np.where(trace == -999.0)
    trace[bad] = 0.0
    return

def applyTrace(xcorr, yfull, trace, tracemask):
    """Apply the trace correction"""
    nevents = len(xcorr)
    ixcorr = xcorr.astype(np.int32)
    #
    # remainder and delta are 1-d arrays
    remainder = xcorr - ixcorr
    delta = (1.0-remainder)*trace[ixcorr] + remainder*trace[ixcorr+1]
    yfull[tracemask] = yfull[tracemask] - delta[tracemask]
    return

def doProfileAlignment(events, input, info, switches, reffiles, phdr, hdr,
                       dq_array, tracemask):
    #
    # First look for a user-defined value for the alignment correction
    # in the SP_SET_A/B keyword
    try:
        segmentletter = info["segment"][-1]
        key = "SP_SET_" + segmentletter
        user_offset = hdr[key]
    except KeyError:
        user_offset = None
    #
    # Now do the profile correction. First we calculate
    # the centroid of the profile
    cosutil.printMsg("Calculating centroid")
    filter = {"segment": info["segment"],
              "opt_elem": info["opt_elem"],
              "cenwave": info["cenwave"],
              "aperture": info["aperture"]}
    #
    #   Use the value of the XTRCTALG keyword to decide what to use to
    #   define regions:
    #   BOXCAR:  Use XTRACTAB
    #   TWOZONE: Use TWOZXTAB
    if info["xtrctalg"] == "BOXCAR":
        xtract_info = cosutil.getTable(reffiles["xtractab"], filter)
        if xtract_info is None:
            raise MissingRowError("Missing row in XTRACTAB; filter = %s" %
                                  str(filter))
    else:
        xtract_info = cosutil.getTable(reffiles["twozxtab"], filter)
        if xtract_info is None:
            raise MissingRowError("Missing row in TWOZXTAB; filter = %s" %
                                  str(filter))
        #
        # Make sure the table doesn't have a SLOPE column
        try:
            slope = xtract_info.field("SLOPE")[0]
            cosutil.printWarning("TWOZXTAB file has a SLOPE column")
        except KeyError:
            slope = 0.0
    #
    # Rebin the data.  Output is the flatfielded image, with events binned
    # in (XFULL, YFULL) space
    rebinned_data = rebinData(events, info)
    nrows, ncols = rebinned_data.shape
    #
    # Mask airglow lines in the DQ image
    maskAirglowLines(dq_array, info, reffiles['disptab'], DQ_AIRGLOW)
    #
    # Apply wavelength limits specified in the PROFTAB reference file
    # so we don't centroid over noise
    applyWavelengthLimits(dq_array, info, reffiles['proftab'],
                          reffiles['disptab'])
    #
    # Determine which columns are good by checking the dq flags in the
    # extraction regions of the background and source
    # The bit value for SDQFLAGS comes from the science data header
    sdqflags = hdr["SDQFLAGS"]
    dqexclude = sdqflags | DQ_AIRGLOW
    centroid = None
    goodcolumns = None
    regions = None
    try:
        slope = xtract_info.field('SLOPE')[0]
    except KeyError:
        slope = 0.0
    startcenter = xtract_info.field('B_SPEC')[0] + slope*(ncols / 2)
    #
    # Calculate the centroid in the science data
    status, centroid, goodcolumns, regions = getScienceCentroid(rebinned_data,
                                                                dq_array,
                                                                xtract_info,
                                                                dqexclude,
                                                                centroid)
    #
    # Calculate the error on the centroid using the counts data which
    # will be calculated from the event list
    error = getCentroidError(events, info, goodcolumns, regions)
    if error is not None:
        cosutil.printMsg("Error on centroid = %f" % (error))
    else:
        cosutil.printWarning("Centroid error not defined in science data")
    #
    # Get the YERRMAX parameter from the twozxtab
    try:
        max_err = xtract_info["yerrmax"][0]
    except KeyError:
        raise MissingColumnError("Missing YERRMAX column in TWOZXTAB, filter = %s" %
                                 str(filter))        
    if error is not None and error > max_err:
        cosutil.printWarning("Error on flux-weighted centroid > %f" % max_err)
        status = CENTROID_ERROR_TOO_LARGE
    ref_centroid = getReferenceCentroid(reffiles['proftab'],
                                        info,
                                        goodcolumns,
                                        regions)
    #
    # Calculate the offset
    if (user_offset is None) or (user_offset < -990.0):
        if status == CENTROID_OK:
            offset = ref_centroid - centroid
            cosutil.printMsg("Using calculated offset of %f" % (offset))
        else:
            #
            # If something went wrong with the centroid calculation, use
            # the center from the xtractab or twozxtab in calculating the
            # offset to make the centroid the same as that in the reference
            # profile
            offset = ref_centroid - startcenter
            cosutil.printMsg("Using offset of {:+f} calculated using".format(offset))
            cosutil.printMsg("initial reference file aperture center")
    else:
        offset = -user_offset
        cosutil.printMsg("Using user-supplied offset of %f" % (offset))
        status = CENTROID_SET_BY_USER
    #
    # Apply the offset.  When this is done, the science target and the
    # reference profile will have the same centroid
    # Only do the correction to events inside the active area and outside
    # the WCA aperture
    applyOffset(events.field('yfull'), offset, tracemask)
    updateTraceKeywords(hdr, info["segment"], ref_centroid, offset, error)
    return status

def rebinData(events, info):
    rebinned = np.zeros(info['npix'], dtype=np.float32)
    ccos.binevents(events.field('xfull'),
                   events.field('yfull'),
                   rebinned,
                   info['x_offset'],
                   events.field('dq'),
                   SERIOUS_DQ_FLAGS,
                   events.field('epsilon'))
    return rebinned

def rebinCounts(events, info):
    rebinnedCounts = np.zeros(info['npix'], dtype=np.float32)
    ccos.binevents(events.field('xfull'),
                   events.field('yfull'),
                   rebinnedCounts,
                   info['x_offset'],
                   events.field('dq'),
                   SERIOUS_DQ_FLAGS)
    return rebinnedCounts

def maskAirglowLines(dq_image, info, disptab, airglow_bits):
    segment = info["segment"]
    for airglow_line in airglow.AIRGLOW_WAVELENGTHS:
        limits = airglow.findAirglowLimits(info, segment,disptab,
                                           airglow_line)
        if limits is not None:
            colstart, colstop = limits
            dq_image[:,int(colstart):int(colstop+1)] = \
            np.bitwise_or(dq_image[:,int(colstart):int(colstop+1)], airglow_bits)
    return

def applyWavelengthLimits(dq_image, info, proftab, disptab):
    """Get the min and max wavelength for our setting.  Calculate
    column numbers for those wavelengths and set dq_image to SERIOUS_DQ_FLAGS
    for columns outside those columns"""
    #
    # Select the row on OPT_ELEM, CENWAVE, SEGMENT and APERTURE
    filter = {'OPT_ELEM': info['opt_elem'],
              'CENWAVE': info['cenwave'],
              'SEGMENT': info['segment'],
              'APERTURE': info['aperture']
              }
    prof_row = cosutil.getTable(proftab, filter)
    if prof_row is None:
        raise MissingRowError("Missing row in PROFTAB: filter = %s" %
                              str(filter))
    wmin, wmax = getWavelengthLimits(prof_row)
    #
    # Now get the dispersion information
    disp_rel = dispersion.Dispersion(disptab, filter)
    min_column = float(disp_rel.evalInvDisp(wmin, tiny=1.0e-8))
    max_column = float(disp_rel.evalInvDisp(wmax, tiny=1.0e-8))
    cosutil.printMsg("Lower wavelength limit of %f corresponds to column %d" % \
                         (wmin, int(min_column)))
    cosutil.printMsg("Upper wavelength limit of %f corresponds to column %d" % \
                         (wmax, int(max_column)))
    if min_column >= 0:
        dq_image[:, 0:int(min_column+1)] = DQ_AIRGLOW
    if max_column <= FUV_X:
        dq_image[:, int(max_column):] = DQ_AIRGLOW
    return

def getWavelengthLimits(prof_row):
    return (900.0, 2100.0)

def getGoodRows(dq_array, dqexclude):
    #
    # Get the first and last good rows in the science data
    nrows, ncols = dq_array.shape
    masked = np.bitwise_and(dq_array, dqexclude)
    mask = np.where(masked != 0, 1, 0)
    rowsum = mask.sum(axis=1)
    #
    # Need to have at least COLMIN good columns
    goodrows = np.where(rowsum < (ncols-COLMIN))
    firstgood = goodrows[0][0]
    lastgood = goodrows[0][-1]
    return firstgood, lastgood

def getRegions(dq_array, xtract_info, dqexclude, center):
    regions = {'bg1start': 0,
               'bg1stop': 0,
               'bg2start': 0,
               'bg2stop': 0,
               'specstart': 0,
               'specstop': 0,
               'firstgoodrow' : 0,
               'lastgoodrow' : 0,
               'center' : 0.0
               }
    #
    # First get the range of rows to use
    regions['firstgoodrow'], regions['lastgoodrow'] = getGoodRows(dq_array,
                                                                  dqexclude)
    nrows, ncols = dq_array.shape
    # BHEIGHT is the width for the background extraction regions
    bgheight = xtract_info["BHEIGHT"][0]
    if center is None:
        # Background and source extraction regions in the XTRACTAB and TWOZXTAB
        # are slanted relative to rows at an angle with tangent = SLOPE.
        # We can assume that the trace takes out the overall slope.
        # The limits of the extraction are calculated at the first column,
        # whereas we need them at the center.  Hence we correct by the
        # delta_y parameter below.
        #
        try:
            slope = xtract_info["SLOPE"][0]
        except KeyError:
            slope = 0.0
        delta_y = slope*ncols/2
        center = xtract_info["B_SPEC"][0] + delta_y
    bg1offset = int(round(xtract_info["B_BKG1"][0] - xtract_info["B_SPEC"][0]))
    regions['bg1start'] = int(round(center)) + bg1offset - bgheight//2
    regions['bg1stop'] = int(round(center)) + bg1offset + bgheight//2
    bg2offset = int(round(xtract_info["B_BKG2"][0] - xtract_info["B_SPEC"][0]))
    regions['bg2start'] = int(round(center)) + bg2offset - bgheight//2
    regions['bg2stop'] = int(round(center)) + bg2offset + bgheight//2
    #
    # Make sure the background regions doesn't extend into bad rows
    regions['bg1start'] = max(regions['bg1start'], regions['firstgoodrow'])
    regions['bg2stop'] = min(regions['bg2stop'], regions['lastgoodrow'])
    specheight = xtract_info["HEIGHT"][0]
    regions['specstart'] = int(round(center)) - specheight//2
    regions['specstop'] = int(round(center)) + specheight//2
    regions['center'] = center
    return regions

def getGoodColumns(dq_array, dqexclude_target, dqexclude_background, regions):
    #
    # Only use columns where the DQ values are good for both background
    # regions and the source
    rowstart = regions['bg1start']
    rowstop = regions['bg1stop']
    dqsum1 = np.bitwise_and(dq_array[int(rowstart):int(rowstop+1)],
                            dqexclude_background).sum(axis=0)

    rowstart = regions['bg2start']
    rowstop = regions['bg2stop']
    dqsum2 = np.bitwise_and(dq_array[int(rowstart):int(rowstop+1)],
                            dqexclude_background).sum(axis=0)

    rowstart = regions['specstart']
    rowstop = regions['specstop']
    dqsum3 = np.bitwise_and(dq_array[int(rowstart):int(rowstop+1)],
                            dqexclude_target).sum(axis=0)

    dqsum = dqsum1 + dqsum2 + dqsum3
    goodcolumns = np.where(dqsum == 0)
    return goodcolumns

def getScienceCentroid(rebinned_data, dq_array, xtract_info,
                       dqexclude, centroid):
    n_iterations = 5
    nrows, ncols = rebinned_data.shape
    center = None
    try:
        slope = xtract_info.field('SLOPE')[0]
    except KeyError:
        slope = 0.0
    startcenter = xtract_info.field('B_SPEC')[0] + slope*(ncols / 2)
    cosutil.printMsg('Starting center = %f' % startcenter)
    for iteration in range(n_iterations):
        #
        # Determine which columns to use.  The rows to be used are centered
        # on the value of the variable 'center', which gets updated between
        # iterations, so the good columns can change from one iteration to
        # the next
        regions = getRegions(dq_array, xtract_info, dqexclude, center)
        #
        # Get the  DQ values for target and background regions.  Both
        # use header value of SDQFLAGS or'd with DQ_AIRGLOW
        # with the gain sag hole DQ taken out.  Put this in a couple of
        # functions so we can change if needed
        dqexclude_target = getTargetDQ(dqexclude)
        dqexclude_background = getBackgroundDQ(dqexclude)
        goodcolumns = getGoodColumns(dq_array, dqexclude_target,
                                     dqexclude_background, regions)
        n_goodcolumns = len(goodcolumns[0])
        cosutil.printMsg("%d good columns" % (len(goodcolumns[0])))
        if n_goodcolumns == 0:
            #
            # If this happens, it means the region over which we are calculating the
            # centroid and associated backgrounds is overlapping part of the detector
            # with non-zero DQ (usually the DQ=8/poorly calibrated regions at the top
            # and bottom edges of the active area)
            # In that case, return with a status of CENTROID_UNDEFINED and set the
            # centroid to the center from the reference file and the goodcolumns and
            # regions to what are appropriate for that centroid
            centroid = startcenter
            center = None
            regions = getRegions(dq_array, xtract_info, dqexclude, center)
            goodcolumns = getGoodColumns(dq_array, dqexclude_target,
                                         dqexclude_background, regions)
            status = CENTROID_UNDEFINED
            cosutil.printMsg("No good columns.")
            cosutil.printMsg("Centroid set to input value from reference file: {}".format(centroid))
            return status, centroid, goodcolumns, regions
        #
        # Now calculate the background.  Use the XTRACTAB reference file
        # to determine the background regions but center them on the 'center'
        # parameter
        background = getBackground(rebinned_data, goodcolumns, regions)
        cosutil.printMsg("Adopted background in science data = %f" %
                         (background))
        rowstart = regions['specstart']
        rowstop = regions['specstop']
        centroid = getCentroid(rebinned_data,
                               goodcolumns,
                               rowstart,
                               rowstop,
                               background)
        if centroid is not None:
            cosutil.printMsg("Centroid = %f" % (centroid))
            difference = (centroid - regions['center'])
            if abs(difference) > 20:
                cosutil.printMsg("Centroid shift too big")
                status = CENTROID_UNDEFINED
                break
            elif abs(difference) < 0.005:
                cosutil.printMsg("Centroid calculation converged")
                status = CENTROID_OK
                break
        else:
            cosutil.printWarning("Centroid is not defined in science data")
            status = CENTROID_UNDEFINED
        center = centroid
        if iteration == n_iterations - 1:
            #
            # If we get here, it means we didn't converge after the maximum
            # number of iterations
            cosutil.printWarning("Centroid calculation did not converge")
            cosutil.printWarning("after %d iterations" % (n_iterations))
            status = NO_CONVERGENCE
            centroid = difference
    return status, centroid, goodcolumns, regions

def getBackgroundDQ(dqexclude):
    """Return the DQ value to be used for the background regions.  For now, this
    means removing the bit corresponding to DQ_GAIN_SAG_HOLE.  To unset a bit,
    AND it with NOT"""
    return dqexclude&(~DQ_GAIN_SAG_HOLE)

def getTargetDQ(dqexclude):
    """Return the DQ value to be used for the target region.  For now, this
    means removing the bit corresponding to DQ_GAIN_SAG_HOLE.  To unset a bit,
    AND it with NOT"""
    return dqexclude&(~DQ_GAIN_SAG_HOLE)

def getBackground(data_array, goodcolumns, regions):
    bg1start = regions['bg1start']
    bg1stop = regions['bg1stop']
    bg2start = regions['bg2start']
    bg2stop = regions['bg2stop']
    cosutil.printMsg("Background regions are from %d to %d" % (bg1start,
                                                               bg1stop))
    cosutil.printMsg("and from %d to %d" % (bg2start, bg2stop))
    bkg1 = data_array[int(bg1start):int(bg1stop+1)].mean(axis=0,
                                                         dtype=np.float64)[goodcolumns].mean(dtype=np.float64)
    nbkg1 = bg1stop - bg1start + 1
    bkg2 = data_array[int(bg2start):int(bg2stop+1)].mean(axis=0,
                                                         dtype=np.float64)[goodcolumns].mean(dtype=np.float64)
    nbkg2 = bg2stop - bg2start + 1
    bkg = (nbkg1*bkg1 + nbkg2*bkg2)/float(nbkg1 + nbkg2)
    return bkg

def getCentroid(data_array, goodcolumns, rowstart, rowstop, background=None):
    """Calculate the centroid of data_array."""
    rows = np.arange(rowstart, rowstop+1)
    if background is None: background = 0.0
    i = data_array[int(rowstart):int(rowstop+1)] - background
    y = rows[:,np.newaxis]
    sumy = (i*y).sum(axis=0, dtype=np.float64)[goodcolumns].sum(dtype=np.float64)
    sumi = i.sum(axis=0,dtype=np.float64)[goodcolumns].sum(dtype=np.float64)
    if sumi != 0.0:
        ycent = sumy/sumi
        return ycent
    else:
        return None

def getCentroidError(events, info, goodcolumns, regions):
    #
    # Calculate the error on the centroid.
    # Use counts data, not flatfielded data.
    # Use the same columns as were used to calculate regular centroid
    # Calculate background for counts data
    # Calculate centroid of counts data
    # Calculate error in centroid for counts data
    # Even though the centroid and background may be different, the error should
    # be appropriate
    #
    if len(goodcolumns[0]) == 0:
        cosutil.printMsg("No good columns, cannot calculate centroid error")
        return None
    # First need to rebin evens to counts
    counts_ij = rebinCounts(events, info)
    #
    # Now need to calculate the background
    background = getBackground(counts_ij, goodcolumns, regions)
    #
    # Calculate centroid
    rowstart = regions['specstart']
    rowstop = regions['specstop']
    centroid = getCentroid(counts_ij, goodcolumns, rowstart, rowstop,
                           background)
    error = calculateCentroidError(counts_ij, goodcolumns, regions,
                                   centroid, background=background)
    return error

def calculateCentroidError(data_ij, goodcolumns, regions,
                           centroid, background=0.0):
    rowstart = regions['specstart']
    rowstop = regions['specstop']
    rows = np.arange(rowstart, rowstop+1)
    if background is None: background = 0.0
    i = data_ij[int(rowstart):int(rowstop+1)]
    y = rows[:,np.newaxis]
    sumi = (i - background).sum(axis=0, dtype=np.float64)[goodcolumns].sum(dtype=np.float64)
    sumsq = (i*(y - centroid) * \
                 (y - centroid)).sum(axis=0,
                                     dtype=np.float64)[goodcolumns].sum(dtype=np.float64)
    if sumi != 0.0:
        error = sumsq / sumi / sumi
        if error >= 0.0:
            return math.sqrt(error)
        else:
            return None
    else:
        return None

def getReferenceBackground(profile, goodcolumns, refcenter, regions):
    nrows, ncols = profile.shape
    if refcenter is None:
        refcenter = nrows/2
    offset = int(round(refcenter - regions['center']))
    bg1start = offset + regions['bg1start']
    bg1stop = offset + regions['bg1stop']
    bg2start = offset + regions['bg2start']
    bg2stop = offset + regions['bg2stop']
    if bg1start >= 0:
        bkg = profile[int(bg1start):int(bg1stop+1)].mean(axis=0,
                                                         dtype=np.float64)[goodcolumns].mean(dtype=np.float64)
        wbkg = 1.0
    elif bg1stop > 0:
        cosutil.printMsg("Background region #1 truncated")
        cosutil.printMsg("Reguested region: %d to %d" % (bg1start, bg1stop))
        wbkg = float(bg1stop+1)/(bg1stop+1-bg1start)
        bg1start = 0
        cosutil.printMsg("Using %d to %d, with weight %f" % (bg1start, 
                                                             bg1stop, wbkg))
        bkg = profile[int(bg1start):int(bg1stop+1)].mean(axis=0,
                                                         dtype=np.float64)[goodcolumns].mean(dtype=np.float64)
    else:
        cosutil.printMsg("Unable to extract background region #1 from reference profile")
        cosutil.printMsg("Requested rows: %d to %d" % (bg1start, bg1stop))
        bkg = 0
        wbkg = 0.0
    if bg2stop <= nrows:
        bkg = bkg + profile[int(bg2start):int(bg2stop+1)].mean(axis=0,
                                                               dtype=np.float64)[goodcolumns].mean(dtype=np.float64)
        wbkg2 = 1.0
    elif bg2start <= nrows:
        cosutil.printMsg("Background region #2 truncated")
        cosutil.printMsg("Reguested region: %d to %d" % (bg2start, bg2stop))
        wbkg2 = float(nrows-bg2start+1)/(bg2stop+1-bg2start)
        bg2stop = nrows
        cosutil.printMsg("Using %d to %d, with weight %f" % (bg2start,
                                                             bg2stop, wbkg2))
        bkg = bkg + profile[int(bg2start):int(bg2stop+1)].mean(axis=0,
                                                               dtype=np.float64)[goodcolumns].mean(dtype=np.float64)
    else:
        cosutil.printMsg("Unable to extract background region #2 from" \
                             " reference profile")
        cosutil.printMsg("Requested rows: %d to %d" % (bg2start, bg2stop))
        wbkg2 = 0.0
    wbkg = wbkg + wbkg2
    if wbkg > 0.0:
        cosutil.printMsg("Adopted background regions for reference profile:")
        cosutil.printMsg("Rows %d to %d, and rows %d to %d" % (bg1start,
                                                               bg1stop,
                                                               bg2start,
                                                               bg2stop))
        background = bkg/wbkg
    else:
        cosutil.printMsg("Unable to determine background in reference profile")
        background = 0.0
    cosutil.printMsg("Background = %f" % background)
    return background

def getReferenceCentroid(proftab, info, goodcolumns, regions):
    """Determine centroid from reference profile

    Parameters
    ----------

    proftab : str
        Name of PROFTAB reference file

    info : dict
        Dictionary of basic info for the exposure

    goodcolumns : slice
        Slice of good columns

    regions : dict
        Dictionary of info about extraction regions

    Returns
    -------

    centroid : float
        Calculated centroid
    """
    #
    # Select the row on OPT_ELEM, CENWAVE, SEGMENT and APERTURE
    filter = {'OPT_ELEM': info['opt_elem'],
              'CENWAVE': info['cenwave'],
              'SEGMENT': info['segment'],
              'APERTURE': info['aperture']
              }
    prof_row = cosutil.getTable(proftab, filter)
    if prof_row is None:
        raise MissingRowError("Missing row in PROFTAB: filter = %s" %
                              str(filter))
    else:
        cosutil.printMsg("Using profile reference file %s" % (proftab))
    #
    # Get the profile and center
    # profile is a 2-d reference spectrum that is used to find out the
    # encircled energy profile for this setting in each column
    profile = prof_row['PROFILE'][0]
    nrows, ncols = profile.shape
    # profile only contains a range of rows, unlike the science data
    # row_0 is where the first row of profile sits in the science data
    row_0 = prof_row['ROW_0'][0]
    # refcenter is the center of the profile calculated using the full
    # range of rows
    refcenter = prof_row['CENTER'][0]
    n_iterations = 5
    for iteration in range(n_iterations):
        # profcenter is the row coordinate of the profile center in the
        # profile array
        profcenter = refcenter - row_0
        cosutil.printMsg("Input Reference centroid = %f, row_0 = %d" %
                         (refcenter, row_0))
        offset = int(round(profcenter - regions['center']))
        rowstart = offset + regions['specstart']
        rowstop = offset + regions['specstop']
        background = getReferenceBackground(profile, goodcolumns,
                                            profcenter, regions)
        centroid = getCentroid(profile, goodcolumns, rowstart, rowstop, 
                               background=background)
        if centroid is None:
            # If getCentroid is unable to determine the centroid, it returns None
            # In that case, set ref_centroid to the value specified in the reference
            # file, in full-height coordinates
            ref_centroid = refcenter
            cosutil.printWarning("Unable to determine centroid in reference spectrum")
            cosutil.printContinuation(f"Using default value of {ref_centroid}")
        else:
            ref_centroid =  centroid + row_0
            cosutil.printMsg(f"Measured reference centroid = {ref_centroid}")
        difference = (refcenter - ref_centroid)
        if abs(difference) < 0.005:
            cosutil.printMsg("Reference centroid calculation converged")
            break
        refcenter = ref_centroid
        if iteration == n_iterations -1:
            #
            # If we get here, it means we didn't converge after the maximum
            # number of iterations
            cosutil.printWarning("Reference centroid calculation did not converge")
            cosutil.printWarning("after %d iterations" % (n_iterations))
    return refcenter

def applyOffset(yfull, offset, tracemask):
    yfull[tracemask] = yfull[tracemask] + offset
    return

def updateTraceKeywords(hdr, segment, ref_centroid, offset, error):
    segmentlist = ['A', 'B']
    segment_letter = segment[-1]
    segmentlist.remove(segment_letter)
    othersegment = segmentlist[0]
    key = "SP_LOC_" + segment_letter
    hdr[key] = ref_centroid
    key = "SP_ERR_" + segment_letter
    if error is not None:
        hdr[key] = error
    else:
        hdr[key] = -999.0
    key = "SP_ERR_" + othersegment
    try:
        temp = hdr[key]
    except KeyError:
        hdr[key] = -999.0
    key = "SP_OFF_" + segment_letter
    hdr[key] = -offset
    return



#! /usr/bin/env python

import os
import sys
import string
import getopt

import numpy as np
from astropy.io import fits

from . import calcosparam                      # parameter definitions
from . import cosutil
from . import getinfo
from . import extract
from . import spwcs
from . import timetag

from .version import *

first_message = True                    # initial values

def main(args):
    """This is a driver to perform 1-D extraction for one file.

    The input is a corrtag file name (the complete name, e.g.
    rootname_corrtag_a.fits, not just a rootname).  One or more input
    files may be specified.
    The suffixes _corrtag, _flt and _counts are assumed by the code,
    so the files on disk must have these suffixes.
    """

    if len(args) < 1:
        print("Specify one or more input corrtag file names.")
        prtOptions()
        raise RuntimeError()

    try:
        (options, pargs) = getopt.getopt(args, "qvuo:",
                                         ["find=", "location=", "extrsize="])
    except Exception as error:
        print(str(error))
        prtOptions()
        raise RuntimeError()

    if len(options) == 0:
        for i in range(len(pargs)):
            if pargs[i][0] == '-':
                prtOptions()
                raise RuntimeError("Command-line options must precede "
                                   "the input file name(s).")

    outdir = None               # output directory name
    update_input = False        # update keywords in corrtag files?
    find_target = {"flag": False, "cutoff": None}
    location = None
    extrsize = None
    for i in range(len(options)):
        if options[i][0] == "-q":
            cosutil.setVerbosity(calcosparam.QUIET)
        elif options[i][0] == "-v":
            cosutil.setVerbosity(calcosparam.VERY_VERBOSE)

        elif options[i][0] == "-u":
            update_input = True

        elif options[i][0] == "-o":
            outdir = options[i][1]

        elif options[i][0] == "--find":
            temp = options[i][1].lower()
            if temp == "yes" or temp == "true":
                find_target["flag"] = True
            elif temp == "no" or temp == "false":
                find_target["flag"] = False
            else:
                try:
                    cutoff = float(temp)
                except ValueError:
                    prtOptions()
                    raise RuntimeError("Don't understand '--find %s'" %
                                       options[i][1])
                if cutoff < 0.:
                    prtOptions()
                    raise RuntimeError("Cutoff for --find cannot be negative.")
                find_target["flag"] = True
                find_target["cutoff"] = cutoff

        elif options[i][0] == "--location":
            values = options[i][1].split()
            location = []
            for i in range(len(values)):
                if values[i].lower() == "none":
                    location.append(None)
                else:
                    location.append(float(values[i]))

        elif options[i][0] == "--extrsize":
            values = options[i][1].split()
            extrsize = []
            for i in range(len(values)):
                if values[i].lower() == "none":
                    extrsize.append(None)
                else:
                    extrsize.append(int(values[i]))

    extractSpec(pargs, outdir, update_input,
                location, extrsize, find_target)

def extractSpec(inlist=[], outdir=None, update_input=False,
                location=None, extrsize=None,
                find_target={"flag": False, "cutoff": None},
                verbosity=None):
    """Extract a 1-D spectrum from each set of flt and counts images.

    The input is a corrtag file name (the complete name, e.g.
    rootname_corrtag_a.fits, not just a rootname).  One or more input
    files may be specified.
    The suffixes _corrtag, _flt and _counts are assumed by the code,
    so the files on disk must have these suffixes.

    If location or extrsize is specified and there is more than one
    name in inlist, note that the same location or extraction size will
    be used for every input file.  For FUV data, only one location or
    extraction size is specified; for NUV data, from one to three values
    may be specified, but these are for the three spectral stripes, not
    for separate input files.

    Parameters
    ----------
    inlist: list of str
        Names of input corrtag files.

    outdir: str or None
        Name of output directory, or None if output files are to be
        written to the default directory.

    update_input: boolean
        If True, update keywords in the input corrtag file or files.

    location: int or float, or list of integers or floats, or None
        The location (or list of three locations for NUV) of the spectrum
        in the cross-dispersion direction, in pixels; this is where the
        spectrum crosses the middle of the detector (index 8192 for FUV,
        512 for NUV).  None means the user did not specify the location.
        If location was specified, that value will be used, regardless
        of the flag in find_target.

    extrsize: int, or list of int, or None
        The height of the extraction box (or list of three heights for NUV)
        in the cross-dispersion direction.  None means the user did not
        specify the extraction height.

    find_target: dictionary
        Keys are "flag" and "cutoff".  If flag is True, this means that we
        should use the location of the target in the cross-dispersion
        direction if the standard deviation (pixels) of the location is
        no larger than cutoff (if cutoff is positive).  flag = False means
        we should use the location determined from the wavecal or as
        specified by the user.  find_target["flag"] will be turned off if
        location is not None.

    verbosity: int {0, 1, 2} or None
        If not None, set verbosity to this level.
    """

    if verbosity is not None:
        if verbosity < 0 or verbosity > 2:
            raise RuntimeError("Verbosity %d is out of range (0, 1, or 2)" %
                               verbosity)
        cosutil.setVerbosity(verbosity)

    cal_ver = calcosparam.CALCOS_VERSION

    if outdir:
        outdir = os.path.expandvars(outdir)
        if not os.path.isdir(outdir):
            raise RuntimeError("The specified output directory "
                               "doesn't exist:  %s" % outdir)
    else:
        outdir = ""

    # Get the names of the input (corrtag), intermediate (flt, counts, x1d_ab),
    # and output (x1d) files.
    filenames = makeFileNames(inlist, outdir)

    # Now check whether any input file is missing or any intermediate or
    # output file already exists.
    missing = checkMissing(filenames)
    already_exists = checkExists(filenames)
    if missing or already_exists:
        raise IOError()

    # keys = filenames.keys()
    # keys.sort()
    keys = sorted(filenames)
    for x1d in keys:
        corrtag_list = filenames[x1d]["corrtag"]
        flt_list = filenames[x1d]["flt"]
        counts_list = filenames[x1d]["counts"]
        x1d_ab_list = filenames[x1d]["x1d_ab"]
        nfiles = len(corrtag_list)
        for i in range(nfiles):
            printFilenames(corrtag_list[i], flt_list[i], counts_list[i],
                           x1d_ab_list[i])
            is_wavecal = makeFltCounts(cal_ver, corrtag_list[i],
                                       flt_list[i], counts_list[i])
            extract.extract1D(flt_list[i], counts_list[i], x1d_ab_list[i],
                              location=location, extrsize=extrsize,
                              find_target=find_target)

        # For FUV, merge the x1d_a.fits and x1d_b.fits files to x1d.fits.
        concatenateSegments(x1d_ab_list, x1d)

        if is_wavecal:
            extract.recomputeWavelengths(x1d)

        # Copy keywords from the x1d file to input files.
        updateSomeKeywords(x1d, filenames, update_input)

def checkMissing(filenames):
    """Check for missing input files.

    Parameters
    ----------
    filenames: dictionary
        Names of input and output files; key is an x1d file name

    Returns
    -------
    list of str
        List of input corrtag files that are not present
    """

    missing = []
    for x1d in filenames:
        corrtag_list = filenames[x1d]["corrtag"]
        for corrtag in corrtag_list:
            if not os.access(corrtag, os.R_OK):
                missing.append(corrtag)
    if missing:
        if len(missing) == 1:
            cosutil.printError("The following input file is missing:")
            cosutil.printContinuation(missing[0])
        else:
            cosutil.printError("The following input files are missing:")
            for corrtag in missing:
                cosutil.printContinuation(corrtag)

    return missing

def checkExists(filenames):
    """Check for output files that already exist.

    Parameters
    ----------
    filenames: dictionary
        Names of input and output files; key is an x1d file name

    Returns
    -------
    list
        List of output files that already exist
    """

    already_exists = []

    keys = sorted(filenames)
    for x1d in keys:
        flt_list = filenames[x1d]["flt"]
        counts_list = filenames[x1d]["counts"]
        x1d_ab_list = filenames[x1d]["x1d_ab"]
        # Separate loops are used here to control the order of names
        # in already_exists.
        for i in range(len(flt_list)):
            if os.access(flt_list[i], os.R_OK):
                already_exists.append(flt_list[i])
        for i in range(len(counts_list)):
            if os.access(counts_list[i], os.R_OK):
                already_exists.append(counts_list[i])
        for i in range(len(x1d_ab_list)):
            if os.access(x1d_ab_list[i], os.R_OK):
                already_exists.append(x1d_ab_list[i])
        if x1d != x1d_ab_list[0]:               # FUV data?
            if os.access(x1d, os.R_OK):
                already_exists.append(x1d)

    if already_exists:
        if len(already_exists) == 1:
            cosutil.printError("The following output file already exists:")
            cosutil.printContinuation(already_exists[0])
        else:
            cosutil.printError("The following output files already exist:")
            for filename in already_exists:
                cosutil.printContinuation(filename)
        cosutil.printError("Output files will not be overwritten.")

    return already_exists

def printFilenames(corrtag, flt, counts, x1d_ab):
    """Print the names of the files."""

    global first_message

    if not cosutil.checkVerbosity(cosutil.VERBOSE):
        return

    if not first_message:
        cosutil.printMsg("")

    names = [("Input", corrtag),
             ("OutFlt", flt),
             ("OutCounts", counts),
             ("x1d", x1d_ab)]
    cosutil.printFilenames(names)
    cosutil.printMsg("")

    first_message = False

def makeFileNames(inlist, outdir=""):
    """Replace suffixes to make the names of all files that we will need.

    The output is the dictionary filenames:
        key is the rootname_x1d.fits file name
        value is a dictionary:
            key is "corrtag", "flt", "counts", or "x1d_ab"
            value is a list of one (NUV) or two (FUV) file names
    For FUV the names in x1d_ab will have suffix "_x1d_a" or "_x1d_b", while
    for NUV the name in x1d_ab will be the same as the key in filenames.

    The flt, counts, x1d_ab, and x1d file names all include the output
    directory name (if any).

    Parameters
    ----------
    inlist: list of str
        List of input corrtag file names

    outdir: str
        Name of the directory for output files, or ""

    Returns
    -------
    filenames: dictionary
        The names of the input and output files
    """

    filenames = {}
    for input in inlist:

        corrtag = os.path.expandvars(input)
        i = corrtag.rfind("corrtag")
        if i < 0:
            raise RuntimeError("File name " + input +
                               " was expected to have suffix 'corrtag'")
        # This is the corrtag file name, but in the output directory.
        outdir_input = os.path.join(outdir, os.path.basename(corrtag))

        flt = replaceSuffix(outdir_input, "_corrtag", "_flt")
        counts = replaceSuffix(outdir_input, "_corrtag", "_counts")
        x1d_ab = replaceSuffix(outdir_input, "_corrtag", "_x1d")
        i = x1d_ab.rfind("_x1d_a")
        if i < 0:
            i = x1d_ab.rfind("_x1d_b")
        if i < 0:                       # NUV file name
            x1d = x1d_ab
        else:                           # FUV file name
            x1d = x1d_ab[:i+4] + x1d_ab[i+6:]   # root_x1d.fits

        if x1d in filenames:
            filenames[x1d]["corrtag"].append(corrtag)
            filenames[x1d]["flt"].append(flt)
            filenames[x1d]["counts"].append(counts)
            filenames[x1d]["x1d_ab"].append(x1d_ab)
        else:
            filenames[x1d] = {
                "corrtag": [corrtag],
                "flt": [flt],
                "counts": [counts],
                "x1d_ab": [x1d_ab]}

    keys = filenames.keys()
    for x1d in keys:
        filenames[x1d]["corrtag"].sort()
        filenames[x1d]["flt"].sort()
        filenames[x1d]["counts"].sort()
        filenames[x1d]["x1d_ab"].sort()

    return filenames

def makeFltCounts(cal_ver, corrtag, flt, counts):
    """Create the flt and counts files.

    Parameters
    ----------
    cal_ver: str
        Calcos version number and date

    corrtag: str
        Name of corrected events table

    flt: str
        Name of effective count rate file

    counts: str
        Name of count rate file

    Returns
    -------
    boolean
        True if the current exposure is a wavecal, based on EXPTYPE
    """

    fd = fits.open(corrtag, mode="copyonwrite")
    phdr = fd[0].header
    phdr["cal_ver"] = cal_ver

    detector = phdr.get("detector")
    headers = timetag.mkHeaders(phdr, fd[1].header)
    exptime_key = cosutil.segmentSpecificKeyword("exptime", phdr["segment"])
    exptime_default = headers[1].get("exptime", default=-1.)
    exptime = headers[1].get(exptime_key, default=exptime_default)

    is_wavecal = phdr["exptype"].find("WAVE") >= 0

    events = fd[1].data
    x = events.field("XFULL")
    y = events.field("YFULL")
    epsilon = events.field("EPSILON")
    dq = events.field("DQ")
    if detector == "FUV":
        npix = (calcosparam.FUV_Y, calcosparam.FUV_EXTENDED_X)
        x_offset = calcosparam.FUV_X_OFFSET
    else:
        npix = (calcosparam.NUV_Y, calcosparam.NUV_EXTENDED_X)
        x_offset = calcosparam.NUV_X_OFFSET

    # Create the data quality array.
    info = getinfo.getGeneralInfo(phdr, headers[1])
    info["corrtag_input"] = True
    switches = getinfo.getSwitchValues(phdr)
    reffiles = getinfo.getRefFileNames(phdr)
    timetag.setActiveArea(events, info, reffiles["brftab"])
    minmax_shift_dict = timetag.getWavecalOffsets(
                events, info, switches["wavecorr"], reffiles["xtractab"],
                reffiles["brftab"])
    dq_array = timetag.doDqicorr(events, corrtag, info, switches, reffiles,
                                 phdr, headers[1], minmax_shift_dict)

    timetag.writeImages(x, y, epsilon, dq,
                        phdr, headers, dq_array, npix, x_offset, exptime,
                        counts, flt)

    updated = False                     # initial value
    wcs = spwcs.SpWcsImage(flt, info, switches["helcorr"],
                           reffiles["spwcstab"], reffiles["xtractab"])
    flag = wcs.writeWCSKeywords()
    if flag:
        updated = True
        cosutil.printMsg("WCS keywords were written for %s" %
                         flt, calcosparam.VERY_VERBOSE)
    wcs = spwcs.SpWcsImage(counts, info, switches["helcorr"],
                           reffiles["spwcstab"], reffiles["xtractab"])
    flag = wcs.writeWCSKeywords()
    if flag:
        updated = True
        cosutil.printMsg("WCS keywords were written for %s" %
                         counts, calcosparam.VERY_VERBOSE)
    if not updated:
        cosutil.printMsg("WCS keywords were not written for the"
                         " flt and counts files.", calcosparam.VERY_VERBOSE)

    fd.close()

    return is_wavecal

def concatenateSegments(x1d_ab_list, x1d):
    """Concatenate the 1-D spectra for the two FUV segments into one file.

    If the input list x1d_ab_list contains a file name that is the same
    as x1d (should be only one name in the list in this case, i.e. NUV
    data), this function will return without doing anything.

    If the list contains just one file name, the file will be renamed;
    otherwise, the files will be concatenated into one file, given by x1d.

    Parameters
    ----------
    x1d_ab_list: list of str
        rootname_x1d_a.fits, rootname_x1d_b.fits file names

    x1d: str
        rootname_x1d.fits file name
    """

    nfiles = len(x1d_ab_list)

    for i in range(nfiles):
        if x1d_ab_list[i] == x1d:
            return

    if nfiles == 1:
        cosutil.renameFile(x1d_ab_list[0], x1d)
    else:
        extract.concatenateFUVSegments(x1d_ab_list, x1d)

def updateSomeKeywords(x1d, filenames, update_input=False):
    """Copy extraction location keywords from x1d.fits to other files.

    Parameters
    ----------
    x1d: str
        Name of the rootname_x1d.fits file

    filenames: dictionary
        Names of input and output files

    update_input: boolean
        True if we should update keywords in the corrtag files
    """

    file_list = []
    x1d_ab = filenames[x1d]["x1d_ab"]
    if len(x1d_ab) > 1:
        file_list.extend(x1d_ab)
    file_list.extend(filenames[x1d]["flt"])
    file_list.extend(filenames[x1d]["counts"])
    if update_input:
        file_list.extend(filenames[x1d]["corrtag"])

    copyKeywords(x1d, file_list)

def copyKeywords(x1d, file_list):
    """Copy extraction location keywords to other headers.

    Parameters
    ----------
    x1d: str
        Name of the x1d.fits file (containing both segments, if FUV)

    file_list: list of str
        Names of the files (e.g. x1d_a.fits, x1d_b.fits) in which header
        keywords should be updated
    """

    fd1 = fits.open(x1d, mode="readonly", memmap=False)

    if fd1[0].header["detector"] == "FUV":
        keywords = ["sp_loc_a", "sp_loc_b",
                    "sp_off_a", "sp_off_b",
                    "sp_nom_a", "sp_nom_b",
                    "sp_slp_a", "sp_slp_b",
                    "sp_hgt_a", "sp_hgt_b",
                    "b_bkg1_a", "b_bkg1_b",
                    "b_bkg2_a", "b_bkg2_b",
                    "b_hgt1_a", "b_hgt1_b",
                    "b_hgt2_a", "b_hgt2_b"]
    else:
        keywords = ["sp_loc_a", "sp_loc_b", "sp_loc_c",
                    "sp_off_a", "sp_off_b", "sp_off_c",
                    "sp_nom_a", "sp_nom_b", "sp_nom_c",
                    "sp_slp_a", "sp_slp_b", "sp_slp_c",
                    "sp_hgt_a", "sp_hgt_b", "sp_hgt_c",
                    "b_bkg1_a", "b_bkg1_b", "b_bkg1_c",
                    "b_bkg2_a", "b_bkg2_b", "b_bkg2_c",
                    "b_hgt1_a", "b_hgt1_b", "b_hgt1_c",
                    "b_hgt2_a", "b_hgt2_b", "b_hgt2_c"]

    for filename in file_list:
        # check that the file still exists (might have been renamed)
        if os.access(filename, os.R_OK):
            fd2 = fits.open(filename, mode="update")
            for key in keywords:
                value = fd1[1].header.get(key, -999.)
                fd2[1].header[key] = value
            fd2.close()

    fd1.close()

def prtOptions():
    """Print a list of command-line options and arguments."""

    print("The command-line arguments and options are:")
    print("  -q (quiet)")
    print("  -v (very verbose)")
    print("  -u (update keywords in input corrtag files)")
    print("  -o outdir (output directory name)")
    print("  --find yes (find Y location of spectrum)")
    print("  --find no (use Y location of spectrum "
                     "from 1dx file and wavecal)")
    print("  --find cutoff (find Y location if sigma <= cutoff)")
    print("  --location Y location(s) at which to extract spectra")
    print("  --extrsize height(s) of extraction region(s)")
    print("  one or more corrtag file names")

def replaceSuffix(rawname, suffix, new_suffix):
    """Replace the suffix in a raw file name.

    Parameters
    ----------
    rawname: str
        A file name

    suffix: str
        Suffix part of rootname to be replaced

    new_suffix: str
        The string to replace suffix in rawname

    Returns
    -------
    string
        rawname with suffix replaced by new_suffix
    """

    lenraw = len(rawname)
    lensuffix = len(suffix)
    i = rawname.rfind(suffix)
    if i >= 0:
        newname = rawname[0:i] + new_suffix + rawname[i+lensuffix:]
    else:
        raise RuntimeError("File name " + rawname +
                           " was expected to have suffix " + suffix)

    return newname

if __name__ == "__main__":

    main(sys.argv[1:])



#!/usr/bin/env python

from setuptools import setup, Extension
from numpy import get_include as numpy_includes
from pathlib import Path


def c_sources(parent: str) -> list[str]:
    return [str(filename) for filename in Path(parent).glob("*.c")]


def c_includes(parent: str, depth: int = 1):
    return [
        parent,
        *(
            str(filename)
            for filename in Path(parent).iterdir()
            if filename.is_dir() and len(filename.parts) - 1 <= depth
        ),
    ]


PACKAGENAME = "calcos"
SOURCES = c_sources("src")
INCLUDES = c_includes("src") + [numpy_includes()]


setup(
    ext_modules=[
        Extension(
            PACKAGENAME + ".ccos",
            sources=SOURCES,
            include_dirs=INCLUDES,
        ),
    ],
)



import copy
import numpy as np
from .calcosparam import *
from . import cosutil

# If there aren't at least this many counts in a wavecal spectrum, flag
# it as not found.
MIN_NUMBER_OF_COUNTS = 50

# For chi square.
N_SIGMA = 15.

# For comparison of individual shifts with the global shift (NUV only).
SLOP = 25.      # pixels

# The maximum number of pixels to set to zero at an end of a spectrum.
TRIM = 10
NUV_X = 1024

# This is used by findShift.  5 is listed last because that's the number
# of points we'll use for determining the shift; the others are for
# finding the scatter of the shift for these values of npts.
NPTS_RANGE = [3, 4, 6, 7, 5]

class Shift1(object):
    """Find the shift in the dispersion direction.

    fs1 = findshift1.Shift1(spectra, templates, info, reffiles,
                            xc_range, fp_pixel_shift, initial_offset=0,
                            spec_found={})
    The public methods are:
        fs1.findShifts()
        shift1 = fs1.getShift1(key)
        fs1.setShift1(key, shift1, fp=True)
        user_specified = fs1.getUserSpecified(key)
        orig_shift1 = fs1.getOrigShift1(key)
        measured_shift1 = getMeasuredShift1(key)
        fp_pixel_shift = getFpPixelShift(key)
        flag = fs1.getSpecFound(key)
        error_estimate = fs1.getScatter(key)
        chi_square = fs1.getChiSq(key)
        number_of_degrees_of_freedom = fs1.getNdf(key)
    The following may be used for testing/debugging:
        spectrum = fs1.getSpec(key)     # a slice, normalized to the template
        template = fs1.getTmpl(key)     # a slice

    Parameters
    ----------
    spectra: dictionary of array_like
        The 1-D extracted spectra; these should be in counts, not counts/s,
        because Poisson statistics will be assumed.  Key is segment or
        stripe name.

    templates: dictionary of array_like
        Template spectra (same keys as for spectra) from the lamptab.
        Key is segment or stripe name.

    info: dictionary
        Header keywords and values.

    reffiles: dictionary
        Reference file names.

    xc_range: int
        The maximum offset (lag) for the cross correlation; this is from
        column XC_RANGE in the WCPTAB.

    fp_pixel_shift: dictionary
        From the FP_PIXEL_SHIFT column of the lamptab, with an entry (same
        keys as for spectra) for each segment or stripe; if that column
        is not present in the lamptab, the values should be 0.

    initial_offset: float
        If the lamptab contains both the FPOFFSET and FP_PIXEL_SHIFT
        columns (this is the normal case), this argument is the value read
        from column SEARCH_OFFSET in the WCPTAB.  If SEARCH_OFFSET is not
        found in the WCPTAB, initial_offset should be set to 0.
        If FPOFFSET and FP_PIXEL_SHIFT are not in the lamptab,
        initial_offset should be SEARCH_OFFSET + (FPOFFSET * STEPSIZE),
        i.e. the sum of SEARCH_OFFSET from WCPTAB and the nominal offset
        between the wavecal spectrum and the template spectrum from the
        lamptab.

    spec_found: dictionary of boolean flags
        True for each spectrum that was found (same keys as for spectra);
        this is a set of initial values which will be saved as an
        attribute, updated, and may be gotten via method getSpecFound.
    """

    def __init__(self, spectra, templates,
                 info, reffiles,
                 xc_range, fp_pixel_shift, initial_offset=0.,
                 spec_found={}):

        self.spectra = copy.deepcopy(spectra)
        self.templates = copy.deepcopy(templates)
        self.info = info
        self.reffiles = reffiles
        self.xc_range = xc_range
        self.lenxc = 2*xc_range + 1
        self.fp_pixel_shift = fp_pixel_shift
        self.initial_offset = initial_offset

        # These are the results.
        self.spec_found = copy.copy(spec_found)         # may be updated
        self.shift1_dict = {}
        self.orig_shift1_dict = {}      # shift1 even if poorly found
        self.user_specified_dict = {}
        self.chisq_dict = {}
        self.ndf_dict = {}              # number of degrees of freedom
        self.scatter_dict = {}
        # for testing; these are aligned slices
        self.spec_dict = {}
        self.tmpl_dict = {}

        # These are for copying info from computeNormalization to
        # computeChiSquare.
        self.factor = None      # spec = baseline + factor * tmpl + noise
        self.baseline = None
        self.rms = None
        self.spec_slice = None
        self.tmpl_slice = None

        # working parameters
        # keys = self.spectra.keys()
        # keys.sort()
        keys = sorted(self.spectra)
        self.keys = keys
        self.current_key = ""

        self.status = 0                 # currently not used

        # LP4 and later wcptab ref files will include the chi-squared threshold
        # Will read from the wcptab column "N_SIGMA" if it exists
        # If not, will revert to default value
        self.chi_sq_threshold = N_SIGMA

        # Trim bright lines off the ends of the spectra.
        self.trimSpectra()

        if not self.spec_found:
            for key in keys:
                self.spec_found[key] = True

        for key in keys:
            self.shift1_dict[key] = 0.
            self.orig_shift1_dict[key] = 0.
            self.user_specified_dict[key] = False
            self.scatter_dict[key] = 0.
            self.chisq_dict[key] = 0.
            self.ndf_dict[key] = 0

    def trimSpectra(self):
        """Trim the ends of the spectra, if bright lines are found.

        This function checks the left and right edges of NUV wavecal spectra,
        looking for emission lines that are partially truncated by the edge.
        For each truncated line that is found, pixel values near the edge will
        be set to zero to reduce or eliminate the effect of the line when
        comparing the wavecal spectrum with the template.
        """

        if self.info["detector"] == "FUV":
            return

        # binning by three pixels and a cutoff of 3*10 are arbitrary, and
        # they may need to be adjusted
        BIN = 3                 # binning of spectrum to get s_l or s_r

        for key in self.keys:
            nelem = len(self.spectra[key])
            index = np.argsort(self.spectra[key])
            median = self.spectra[key][index[nelem//2]]
            cutoff_brightness = float(BIN * max(1., median))
            cutoff_brightness = max(BIN*10., cutoff_brightness)
            # first and last pixels that are on the detector
            left = int(self.info["x_offset"])
            right = left + NUV_X - 1
            assert left < nelem and right < nelem
            # first and last twelve (detector) pixel values, binned by three
            # (so these are four-element arrays)
            s_l = self.spectra[key][left:left+12:BIN] + \
                  self.spectra[key][left+1:left+13:BIN] + \
                  self.spectra[key][left+2:left+14:BIN]
            s_r = self.spectra[key][right-11:right-1:BIN] + \
                  self.spectra[key][right-10:right:BIN] + \
                  self.spectra[key][right-9:right+1:BIN]
            # mod_left and mod_right are flags that can be set within a loop;
            # if True, a message will be printed (outside the loop) to say
            # that a wavecal spectral line was on the left or right edge and
            # has been clobbered.
            mod_left = False
            mod_right = False
            # Look for values in the binned arrays that are bright (compared
            # with the median) and decreasing in brightness away from the edge.
            if s_l[0] > cutoff_brightness:
                i = 0               # index in s_l
                j = left            # index in spectrum
                while i < 3:
                    if s_l[i] > s_l[i+1]:
                        mod_left = True
                        self.spectra[key][j:j+BIN] = 0.
                        i += 1; j += BIN
                    else:
                        break
            if s_r[3] > cutoff_brightness:
                i = 3; j = right + 1
                while i >= 0:
                    if s_r[i] > s_r[i-1]:
                        mod_right = True
                        self.spectra[key][j-BIN:j] = 0.
                        i -= 1; j -= BIN
                    else:
                        break

            if mod_left:
                cosutil.printMsg("Info:  truncated line removed at left"
                                 " edge of %s" % key)
            if mod_right:
                cosutil.printMsg("Info:  truncated line removed at right"
                                 " edge of %s" % key)

    def getShift1(self, key):
        """Return the shift in the dispersion direction."""

        if key in self.shift1_dict:
            return self.shift1_dict[key] + self.fp_pixel_shift[key]
        else:
            return 0.

    def setShift1(self, key, shift1, fp=True):
        """Set shift1 to the value supplied by the user."""

        if key not in self.shift1_dict:
            return

        spectrum = self.spectra[key]
        template = self.templates[key]
        self.computeNormalization(spectrum, template,
                                  shift1 - self.fp_pixel_shift[key])
        if self.factor is None:
            self.chisq_dict[key] = 0.
            self.ndf_dict[key] = 0
            self.spec_dict[key] = None
            self.tmpl_dict[key] = None
        else:
            (chisq, ndf, spec, tmpl) = \
                        self.computeChiSquare(spectrum, template)
            self.chisq_dict[key] = chisq
            self.ndf_dict[key] = ndf
            self.spec_dict[key] = spec.copy()
            self.tmpl_dict[key] = tmpl.copy()

        # the default is to include the fp_pixel_shift offset
        if fp:
            self.shift1_dict[key] = shift1 - self.fp_pixel_shift[key]
        else:
            self.shift1_dict[key] = shift1
        self.spec_found[key] = True
        self.user_specified_dict[key] = True
        self.scatter_dict[key] = 0.

    def getUserSpecified(self, key):
        """Return True if shift1 was specified by the user."""

        if key in self.user_specified_dict:
            return self.user_specified_dict[key]
        else:
            return False

    def getOrigShift1(self, key):
        """Return the shift1 value even if it was poorly found."""

        if key in self.shift1_dict:
            return self.orig_shift1_dict[key] + self.fp_pixel_shift[key]
        else:
            return 0.

    def getMeasuredShift1(self, key):
        """Return the shift1 value that was directly measured."""

        if key in self.shift1_dict:
            return self.orig_shift1_dict[key]
        else:
            return 0.

    def getFpPixelShift(self, key):
        """Return the fp_pixel_shift value."""

        if key in self.shift1_dict:
            return self.fp_pixel_shift[key]
        else:
            return 0.

    def getSpecFound(self, key):
        """Return a flag indicating whether the spectrum was found."""

        if key in self.spec_found:
            return self.spec_found[key]
        else:
            return False

    def getScatter(self, key):
        """Return an estimate of the uncertainty in shift1."""

        if key in self.scatter_dict:
            return self.scatter_dict[key]
        else:
            return 0.

    def getChiSq(self, key):
        """Return Chi square for the spectrum vs template."""

        if key in self.chisq_dict:
            return self.chisq_dict[key]
        else:
            return -1.

    def getNdf(self, key):
        """Return the number of number of degrees of freedom for Chi square."""

        if key in self.ndf_dict:
            return self.ndf_dict[key]
        else:
            return 0

    def getSpec(self, key):
        """Return the extracted spectrum for testing."""

        if key in self.spec_dict:
            return self.spec_dict[key]
        else:
            return np.zeros(1, dtype=np.float32)

    def getTmpl(self, key):
        """Return the template spectrum for testing."""

        if key in self.tmpl_dict:
            return self.tmpl_dict[key]
        else:
            return np.zeros(1, dtype=np.float32)

    def findShifts(self):
        """Find the shifts in the dispersion direction.

        This function updates:
            self.spec_found
            self.shift1_dict
            self.scatter_dict
            self.chisq_dict
            self.ndf_dict
            self.spec_dict
            self.tmpl_dict
        """

        nelem = len(self.keys)
        if nelem < 1:
            return
        #
        # Try reading the chi square threshold from the wcptab
        opt_elem = self.info["opt_elem"]
        wcp_info = cosutil.getTable(self.reffiles["wcptab"],
                                    filter={"opt_elem": opt_elem},
                                    exactly_one=True)
        try:
            # This field should exist in LP4 and later data
            self.chi_sq_threshold = wcp_info.field("N_SIGMA")
        except KeyError:
            # This could occur if the wcptab doesn't have the
            # N_SIGMA column.
            # We already have the threshold set in the constructor, so
            # we don't need to do anything if we can't get it from the
            # wcptab
            pass
        self.checkCounts()              # flag spectra with negligible counts

        if self.info["detector"] == "FUV":
            self.findShiftsFUV()
        else:
            self.findShiftsNUV()

    def findShiftsFUV(self):
        """Find the shifts in the dispersion direction for FUV data."""

        for key in self.keys:
            if key not in self.templates.keys():
                self.notFound(key)
                continue
            self.current_key = key
            spectrum = self.spectra[key]
            template = self.templates[key]
            (shift, orig_shift1, scatter, foundit) = \
                        self.findShift(spectrum, template)
            self.orig_shift1_dict[key] = orig_shift1
            self.scatter_dict[key] = scatter
            if not foundit:
                self.notFound(key)
                continue

            self.computeNormalization(spectrum, template, shift)
            if self.factor is None:
                self.notFound(key)
                continue
            (chisq, ndf, spec, tmpl) = \
                        self.computeChiSquare(spectrum, template)
            self.chisq_dict[key] = chisq
            self.ndf_dict[key] = ndf
            self.spec_dict[key] = spec.copy()
            self.tmpl_dict[key] = tmpl.copy()
            if ndf > 0:
                ratio = chisq / ndf
            else:
                ratio = chisq
            if shift is None or ratio > self.chi_sq_threshold \
                    or ratio < 1./self.chi_sq_threshold:
                shift = 0.
                self.spec_found[key] = False
            self.shift1_dict[key] = shift

    def findShiftsNUV(self):
        """Find the shifts in the dispersion direction for NUV data."""

        global_shift = self.globalShift()

        for key in self.keys:
            if key not in self.templates.keys():
                self.notFound(key)
                continue
            self.current_key = key
            spectrum = self.spectra[key]
            template = self.templates[key]
            (shift, orig_shift1, scatter, foundit) = \
                        self.findShift(spectrum, template)
            self.orig_shift1_dict[key] = orig_shift1
            self.scatter_dict[key] = scatter
            if not foundit:
                self.spec_found[key] = False
            elif global_shift is not None:
                if abs(shift - global_shift) > SLOP:
                    self.spec_found[key] = False
            self.shift1_dict[key] = shift

            self.computeNormalization(spectrum, template, shift)
            if self.factor is None:
                self.notFound(key)
                continue

            (chisq, ndf, spec, tmpl) = \
                        self.computeChiSquare(spectrum, template)
            self.chisq_dict[key] = chisq
            self.ndf_dict[key] = ndf
            self.spec_dict[key] = spec.copy()
            self.tmpl_dict[key] = tmpl.copy()
            if ndf > 0:
                ratio = chisq / ndf
            else:
                ratio = chisq
            if shift is None or ratio > self.chi_sq_threshold \
                    or ratio < 1./self.chi_sq_threshold:
                self.spec_found[key] = False
                shift = 0.

        self.repairNUV()        # assign best-guess values for bad shifts

    def notFound(self, key):
        self.shift1_dict[key] = 0.
        self.spec_found[key] = False
        self.chisq_dict[key] = 0.
        self.ndf_dict[key] = 0

    def checkCounts(self):
        """Flag data with negligible counts.

        This function updates:
            self.spec_found
        """

        for key in self.keys:
            self.current_key = key
            if self.spectra[key].sum() < MIN_NUMBER_OF_COUNTS:
                self.spec_found[key] = False

    def globalShift(self):
        """Return the shift of the sum of all NUV stripes.

        Returns
        -------
        float
            The shift of the sum of all (nominally three) NUV stripes
        """

        key = self.keys[0]

        # Add spectra together, add templates together, find the shift.
        nelem = len(self.spectra[key])
        sum_spectra = np.zeros(nelem, dtype=np.float64)
        sum_templates = np.zeros(nelem, dtype=np.float64)
        nsum = 0
        for key in self.keys:
            self.current_key = key
            # Skip spectra for which spec_found is already set to False,
            # because they probably have negligible counts.
            if key in self.templates and self.spec_found[key]:
                sum_spectra += self.spectra[key]
                sum_templates += self.templates[key]
                nsum += 1
        if nsum < 1:
            return 0

        self.current_key = "all"
        (global_shift, orig_shift1, scatter, foundit) = \
                self.findShift(sum_spectra, sum_templates)
        if not foundit:
            global_shift = None

        return global_shift

    def findShift(self, spectrum, template):
        """Find a shift in the dispersion direction.

        Parameters
        ----------
        spectrum: array_like
            A 1-D extracted wavecal spectrum

        template: array_like
            Template spectrum

        Returns
        -------
        tuple
            (shift, orig_shift1, scatter, foundit), where shift is the
            shift in the dispersion direction, orig_shift1 is the shift
            even if it wasn't well determined, scatter is an estimate of
            the uncertainty in shift1, and foundit is True if we think the
            shift was actually found
        """

        # xxx need to improve the comments in this function

        GOOD_VALUE = 1
        BAD_VALUE = -1

        lenxc = self.lenxc
        # These arrays are for finding a minimum or maximum.
        # We expect to find a minimum rms.
        # These arrays all have length self.lenxc, and we can
        # use the same index(e.g. imin) in all of these arrays.
        rms = np.zeros(lenxc, dtype=np.float32)
        chisq = np.zeros(lenxc, dtype=np.float32)
        factor = np.zeros(lenxc, dtype=np.float32)
        baseline = np.zeros(lenxc, dtype=np.float32)
        flag = np.zeros(lenxc, dtype=np.int32)
        maxlag = lenxc // 2
        i = 0
        # Assign reduced chi square to chisq.
        for shift in range(-maxlag, maxlag+1):
            shift_x = shift + self.initial_offset
            # compute self.rms and other values
            self.computeNormalization(spectrum, template, shift_x)
            if self.factor is not None:
                (chisq_i, ndf, spec, tmpl) = \
                        self.computeChiSquare(spectrum, template)
                chisq[i] = chisq_i / float(max(ndf, 1))
            if self.factor is None:
                flag[i] = BAD_VALUE
            else:
                flag[i] = GOOD_VALUE
                rms[i] = self.rms
            i += 1
        # Where factor was None, set chisq to a large value.
        max_xc = chisq.max()
        chisq = np.where(flag == GOOD_VALUE, chisq, 2.*max_xc)

        good_values = np.where(flag == GOOD_VALUE)
        # Find all the local minima in the array of RMS values.
        n = lenxc
        # indices of local minima in rms
        local_minima = np.where(np.logical_and(rms[1:n-1] <= rms[0:n-2],
                                               rms[1:n-1] <= rms[2:n]))
        if len(good_values[0]) <= 0 or len(local_minima[0]) <= 0:
            return (0., 0., 0., False)
        # Extract the array of indices, and add one to get indices in rms
        # (because rms[1:n-1] starts with 1).
        local_minima = local_minima[0] + 1

        # Check each local minimum to make sure none of the nearby values
        # is flagged as bad.
        # real_minima, rms_list, and chisq_list will all have the same
        # number of elements.
        real_minima = []        # list of indices (imin) of minima of rms
        chisq_list = []         # chi square at each imin in real_minima
        rms_list = []           # value of rms at each imin in real_minima
        npts = max(NPTS_RANGE)
        for imin in local_minima:
            i1 = imin - npts//2
            i1 = max(i1, 0)
            i2 = i1 + npts
            i2 = min(i2, lenxc)
            i1 = i2 - npts
            bad = (flag[i1:i2] == BAD_VALUE)
            if np.any(bad):
                continue
            real_minima.append(imin)
            rms_list.append(rms[imin])
            chisq_list.append(chisq[imin])
        if len(real_minima) <= 0:
            return (0., 0., 0., False)

        # Pick the location with the smallest RMS.  index_of_min is the
        # index in rms_list that gives the smallest RMS.
        min_rms = None
        index_of_min = 0                # initial values
        for (i, rms_i) in enumerate(rms_list):
            chisq_i = chisq_list[i]

        # The values in rms_index and chisq_index are indices in real_minima,
        # rms_list, and chisq_list.
        # index_of_min (below) will also be an index in real_minima, etc.
        # Example:
        # k = rms_index[0]         the point with the minimum RMS
        # imin = real_minima[k]    the index in rms, factor, baseline, flag
        # rms[imin] and rms_list[k] will be the same
        rms_array = np.array(rms_list)
        rms_index = np.argsort(rms_array)
        chisq_array = np.array(chisq_list)
        chisq_index = np.argsort(chisq_array)

        # Of the PICK_N points with the smallest RMS, skip ones that don't have
        # positive curvature, then select the one with smallest chi square.
        PICK_N = min(10, len(rms_index))
        x = np.arange(5, dtype=np.float64)      # for fitting a quadratic
        index_of_min = None                     # initial value
        for i in range(PICK_N):
            # k is an index in real_minima, rms_list, and chisq_list, while
            # real_minima[k] is an index in rms.
            k = rms_index[i]
            imin = real_minima[k]
            # fit a quadratic to five points centered on imin, to get the
            # curvature
            j1 = imin - 2
            j1 = max(j1, 0)
            j2 = j1 + 5
            j2 = min(j2, lenxc)
            j1 = j2 - 5
            (coeff, var) = cosutil.fitQuadratic(x, rms[j1:j2])
            if coeff is None or coeff[2] <= 0.: # coeff[2] is the curvature
                continue
            if index_of_min is None or chisq_list[k] < min_chisq:
                min_chisq = chisq_list[k]
                index_of_min = k
        if index_of_min is None:
            return (0., 0., 0., False)

        # Fit a quadratic to points near the minimum of chisq.
        imin = real_minima[index_of_min]
        (shift, orig_shift1, scatter) = self.findMinimum(imin, maxlag, chisq)

        return (shift, orig_shift1, scatter, True)

    def findMinimum(self, imin, maxlag, chisq):
        """Fit a quadratic to points near the minimum of chisq.

        Parameters
        ----------
        imin: int
            Index to use as a starting point

        maxlag: int
            Half the search range, i.e. maximum offset from nominal

        chisq: array_like
            Reduced chi square, one for each offset

        Returns
        -------
        tuple of three floats
            (shift, orig_shift1, scatter), where shift is the shift in the
            dispersion direction (or zero if not found), orig_shift1 will
            be the same value if the shift was found (but will be the shift
            based on imin if not found), and scatter is a measure of the
            uncertainty in the shift
        """

        lenxc = self.lenxc
        min_shift = None
        max_shift = None
        for npts in NPTS_RANGE:
            i1 = imin - npts//2
            i1 = max(i1, 0)
            i2 = i1 + npts
            i2 = min(i2, lenxc)
            i1 = i2 - npts
            x = np.arange(npts, dtype=np.float64)
            (coeff, var) = cosutil.fitQuadratic(x, chisq[i1:i2])
            (x_min, sigma_shift) = cosutil.centerOfQuadratic(coeff, var)
            if x_min is None:
                shift = 0.
                orig_shift1 = imin + self.initial_offset - maxlag
                self.spec_found[self.current_key] = False
            else:
                shift = x_min + i1 + self.initial_offset - maxlag
                orig_shift1 = shift
                if min_shift is None or shift < min_shift:
                    min_shift = shift
                if max_shift is None or shift > max_shift:
                    max_shift = shift
        if min_shift is None or max_shift is None:
            scatter = sigma_shift
        else:
            scatter = max(sigma_shift, (max_shift - min_shift) / 2.)

        return (shift, orig_shift1, scatter)

    def computeNormalization(self, spectrum, template, shift):
        """Compute a normalization factor between spectrum and template.

        The following attributes are assigned, to be used by computeChiSquare:
            self.factor
            self.baseline
            self.rms
                where spectrum = self.baseline + self.factor * template + noise
                within an overlap region (see spec_slice and tmpl_slice)
            self.spec_slice: the slice of the spectrum to use
            self.tmpl_slice: the slice of the template to use

        self.factor will be set to None if the normalization could not
        be found due to one of these conditions:
            no non-zero data (overlap region could not be found)
            singularity in the fit between template and spectrum
            factor is zero or negative

        Parameters
        ----------
        spectrum: array_like
            The 1-D extracted wavecal spectrum

        template: array_like
            Template spectrum

        shift: float or None
            The pixel shift in the dispersion direction, or None if the
            shift was not found successfully
        """

        self.factor = None
        self.baseline = 0.
        self.rms = 0.

        if shift is None:
            shift = 0.
        shift = int(round(shift))

        len_spec = len(spectrum)

        # Get the overlap region.
        if shift >= 0:
            s0 = shift
            s1 = len_spec
            t0 = 0
            t1 = len_spec - shift
        else:
            s0 = 0
            s1 = len_spec - (-shift)
            t0 = -shift
            t1 = len_spec

        # Narrow the endpoints to exclude elements that are zero in either
        # spectrum or template.
        done = False
        while not done:
            if spectrum[s0] != 0. and template[t0] != 0.:
                break
            s0 += 1
            t0 += 1
            if s0 >= s1-1:
                done = True
        while not done:
            if spectrum[s1-1] != 0. and template[t1-1] != 0.:
                break
            s1 -= 1
            t1 -= 1
            if s1 <= s0:
                done = True
        if done:
            self.factor = None
            return

        self.spec_slice = (s0, s1)
        self.tmpl_slice = (t0, t1)

        # Fit the template to the spectrum:
        # spec = baseline + factor * tmpl + noise
        n = float(s1 - s0)
        spec = spectrum[s0:s1]
        tmpl = template[t0:t1]
        sum_s = spec.sum(dtype=np.float64)
        sum_t = tmpl.sum(dtype=np.float64)
        sum_t2 = (tmpl**2).sum(dtype=np.float64)
        sum_st = (spec * tmpl).sum(dtype=np.float64)
        denominator = sum_t**2 - n * sum_t2
        if denominator == 0.:
            self.factor = None
            return
        else:
            self.factor = (sum_s * sum_t - n * sum_st) / denominator
            if self.factor <= 0.:
                self.factor = None
                return

        self.baseline = (sum_s - self.factor * sum_t) / n

        diff = spec - (self.baseline + self.factor * tmpl)
        nelem = len(diff)
        if nelem > 1:
            self.rms = float(np.sqrt((diff**2).sum() / (nelem-1.)))

    def computeChiSquare(self, spectrum, template):
        """Compute chi square for spectrum and template.

        This method makes use of the following attributes that were
        assigned by computeNormalization:
            self.factor
            self.baseline
            self.spec_slice
            self.tmpl_slice

        Parameters
        ----------
        spectrum: array_like
            The 1-D extracted wavecal spectrum

        template: array_like
            Template spectrum

        Returns
        -------
        tuple
            Chi square, the number of degrees of freedom, the overlapping
            slice of the spectrum and normalized template
        """

        (s0, s1) = self.spec_slice
        (t0, t1) = self.tmpl_slice
        spec = spectrum[s0:s1]
        tmpl = template[t0:t1]

        # Normalize the spectrum to match the template.
        n_spec = (spec - self.baseline) / self.factor

        # sigma for the template = sqrt(template);
        # sigma for the normalized spectrum = sqrt(spectrum) / factor;
        # variance for the normalized spectrum = spectrum / factor**2
        # add variances (add sigmas in quadrature)
        variance = spec / self.factor**2 + tmpl

        # When computing chi square, include only those elements for which
        # either the spectrum or the template is non-zero.
        # (see both_zero below)
        either_positive = np.logical_or(spec > 0., tmpl > 0.)
        nelem = either_positive.sum(dtype=np.float64)
        ndf = max(0, nelem - 1)                 # number of degrees of freedom
        # v is scratch, just so we can divide by it
        v = np.where(variance > 0., variance, 1.)

        # Compute chi square.
        diff = (n_spec - tmpl)
        a_chisq = diff**2 / v                   # array of values
        # Truncate chi square at 1 if both the spectrum and template are
        # 0 or 1.  This cutoff is pretty arbitrary, just a number that's
        # clearly in the noise.
        both_small = np.logical_and(spec <= 1.5, tmpl <= 1.5)
        a_chisq = np.where(both_small, np.minimum(a_chisq, 1.), a_chisq)
        # If both the spectrum (not scaled) and template are zero,
        # chi square should be zero (regardless of scaling).
        both_zero = np.logical_and(spec == 0., tmpl == 0.)
        a_chisq = np.where(both_zero, 0., a_chisq)

        chisq = float(a_chisq.sum(dtype=np.float64))

        return (chisq, ndf, n_spec, tmpl)

    def repairNUV(self):
        """Assign reasonable values for shifts that weren't found."""

        # This is an estimate of the relative shifts between stripes.
        # (52 pixels per fpoffset step)
        # These polynomial coefficients should be read from a reference table.
        offset = {("NUVA", "G185M"): [-0.5, 0.02307692],
                  ("NUVB", "G185M"): [0.0, 0.0],
                  ("NUVC", "G185M"): [1.0, 0.02307692],
                  ("NUVA", "G225M"): [-0.5, 0.02307692],
                  ("NUVB", "G225M"): [0.0, 0.0],
                  ("NUVC", "G225M"): [1.0, 0.02307692],
                  ("NUVA", "G285M"): [-0.5, 0.02307692],
                  ("NUVB", "G285M"): [0.0, 0.0],
                  ("NUVC", "G285M"): [1.0, 0.02307692],
                  ("NUVA", "G230L"): [0.0, 0.0],
                  ("NUVB", "G230L"): [0.0, 0.0],
                  ("NUVC", "G230L"): [0.0, 0.0]}
        # note:
        # 0.02307692 = 1.2 / 52, i.e. 1.2 pixels per fpoffset step of 52 pixels

        # Compute the average shift of the stripes for which shift1 was
        # found, first applying the offset between the current stripe and
        # NUVB.  The average will then be of NUVB-equivalent shifts.

        # shift_0 (in the loop below) is the shift of the wavecal spectrum
        # from the template at fpoffset = 0.  This is used under the assumption
        # that the relative offsets between stripes depends on this shift.
        # Currently, shift_0 is just the anticipated fpoffset shift, not
        # the total shift1[abc] value.
        # shift_to_nuvb is the relative shift from the current stripe to
        # NUVB, based on shift_0 and the dictionary 'offset'.

        ngood = 0
        sum_shifts = 0.
        for key in self.keys:
            self.current_key = key
            off_key = (key, self.info["opt_elem"])
            if self.spec_found[key]:
                shift_0 = self.initial_offset + self.fp_pixel_shift[key]
                shift_to_nuvb = self.evalPoly(shift_0, offset[off_key])
                sum_shifts += (self.shift1_dict[key] - shift_to_nuvb)
                ngood += 1
        if ngood == 0:
            return              # no data; can't do anything

        # Now use the average NUVB-equivalent shift to replace the shift of
        # any stripe that wasn't found, applying the offset from NUVB to the
        # missing stripe or stripes.
        mean_shift = sum_shifts / ngood
        for key in self.keys:
            self.current_key = key
            off_key = (key, self.info["opt_elem"])
            if not self.spec_found[key]:
                shift_0 = self.initial_offset + self.fp_pixel_shift[key]
                shift_to_nuvb = self.evalPoly(shift_0, offset[off_key])
                self.shift1_dict[key] = mean_shift + shift_to_nuvb

    def evalPoly(self, x, coeff):
        """Evaluate a polynomial in x.

        Parameters
        ----------
        x: float
            Argument

        Returns
        -------
        float
            coeff[0] + coeff[1] * x + coeff[2] * x**2 + ...
        """

        ncoeff = len(coeff)
        sum = coeff[ncoeff-1]
        for i in range(ncoeff-2, -1, -1):
            sum = sum * x + coeff[i]
        return sum
